library(raster)
library(sp)
library(gstat)
library(dplyr)
library(stringr)
library(ncdf4)
library(RNetCDF)
library(purrr)
library(automap)

##############################################################
#
#   PARAMETRI
#
##############################################################

# outputDir <- "output"
projutm <- "+proj=utm +zone=32 +datum=WGS84"
projll <- "+proj=longlat +datum=WGS84"

grigliaUTM <- raster(xmn=440000, xmx=720000, ymn=4924000, ymx=5202000, resolution=c(2000,2000), crs=projutm)
grigliaLL <- raster(xmn=8.38, xmx=11.64, ymn=44.58, ymx=46.84, resolution=c(0.02,0.02), crs=projll)


# ONDEMAND

timestepOndemand <- 1 # hours
NTCNameOndemand <- "kriging_ondemand.nc"
NTCVariablesListOndemand <- c("CUM")
UnitsListOndemand <- rep("mm",length(NTCVariablesListOndemand))
LongNameListOndemand <- c("Pioggia cumulata")


##############################################################
#
#   FUNZIONI CREAZIONE E LETTURA NETCDF
#
##############################################################

createNetCDF_LatLon <- function(ncdName, x.vector, y.vector, var.list, unit.list, longName.list) {
  
  time_unit <- "hours since 1900-01-01 00:00:00"
  strTime_unit <- "YYYY-MM-DD hh:mm:ss"
  
  xdim <- ncdim_def("x",units="degrees_east", longname="longitude",as.double(x.vector))
  ydim <- ncdim_def("y",units="degrees_north",longname="latitude",as.double(y.vector))
  timedim <- ncdim_def("time",time_unit,0, unlim=TRUE)
  dimnchar   <- ncdim_def("nchar","", 1:19, create_dimvar=FALSE )
  
  listVar <- vector(mode = "list", length = length(var.list) + 1)
  #ADD NEW VARIABLES TO THE FILE
  for(i in seq(from=1, to=length(var.list), by=1)) {
    
    listVar[[i]] <- ncvar_def(var.list[i],unit.list[i],list(xdim,ydim,timedim),NA,longName.list[i],prec="double", compression = 9)
    # listVar[[i]] <- ncvar_def(var.list[i],unit.list[i],list(xdim,ydim,timedim),NA,longName.list[i],prec="double")
    
  }
  
  # listVar[[i+1]] <- ncvar_def("UTM_Projection","1",NULL,NULL,longname="UTM_Projection",prec="char")
  listVar[[i+1]] <- ncvar_def("strTime",strTime_unit,list(dimnchar,timedim),NULL,longname="Time in format: YYYY-MM-DD hh:mm:ss",prec="char")
  
  ncout <- nc_create(ncdName, listVar, force_v4=TRUE)
  
  # put additional attributes into dimension and data variables
  # ncatt_put(ncout,"x","axis","X")
  # ncatt_put(ncout,"x","standard_name","longitude")
  # ncatt_put(ncout,"x","_CoordinateAxisType","GeoX")
  ncvar_put(ncout,"x",x.vector)
  
  # ncatt_put(ncout,"y","axis","Y")
  # ncatt_put(ncout,"y","standard_name","latitude")
  # ncatt_put(ncout,"y","_CoordinateAxisType","GeoY")
  ncvar_put(ncout,"y",y.vector)
  
  ncatt_put(ncout,"time","_CoordinateAxisType","Time")
  ncatt_put(ncout,"time", "calendar", "gregorian")
  
  cellsize <-  x.vector[2] -  x.vector[1]
  xllcenter <- min(x.vector)
  yllcenter <- min(y.vector)
  xorig <- xllcenter - cellsize/2
  yorig <- yllcenter - cellsize/2
  
  # add global attributes
  ncatt_put(ncout,0,"NCOLS",length(x.vector), prec="double")
  ncatt_put(ncout,0,"NROWS",length(y.vector), prec="double")
  ncatt_put(ncout,0,"NVARS",length(var.list), prec="double")
  ncatt_put(ncout,0,"XLLCORNER",xorig, prec="double")
  ncatt_put(ncout,0,"YLLCORNER",yorig, prec="double")
  ncatt_put(ncout,0,"XLLCENTER",xllcenter, prec="double")
  ncatt_put(ncout,0,"YLLCENTER",yllcenter, prec="double")
  ncatt_put(ncout,0,"CELLSIZE",cellsize, prec="double")
  ncatt_put(ncout,0,"VARNAMES",paste(var.list,collapse=","), prec="text")
  ncatt_put(ncout,0,"EPSG",4326, prec="short")
  ncatt_put(ncout,0,"USED_TIMEZONE","GMT+01:00", prec="text")
  ncatt_put(ncout,0,"Conventions","CF-1.6", prec="text")
  
  nc_close(ncout)
  
}

duplicateNetCDF_LatLon <- function(ncdName,ncdNameOld,timeIndexINIT,timeIndexEND) {
  
  nc <- nc_open( ncdNameOld, readunlim=FALSE )
  
  x.vector <- nc$dim$x$vals
  y.vector <- nc$dim$y$vals
  var.list <- unlist(map(nc$var, "name"), use.names = F)[1:(nc$nvars-1)]
  longName.list <- unlist(map(nc$var, "longname"), use.names = F)[1:(nc$nvars-1)]
  unit.list <- unlist(map(nc$var, "units"), use.names = F)[1:(nc$nvars-1)]
  
  # if (file.exists(ncdName)) file.remove(ncdName)
  
  createNetCDF_LatLon(ncdName,x.vector,y.vector,var.list,unit.list,longName.list)
  
  for (var in var.list) {
    
    timeToCopy <- readVarNetCDF(ncdNameOld,"time")[timeIndexINIT:timeIndexEND]
    strTimeToCopy <- readVarNetCDF(ncdNameOld,"strTime")[timeIndexINIT:timeIndexEND]
    varTocopy <- readVarNetCDF(ncdNameOld,var)
    
    if (length(dim(varTocopy)) == 3) {
      varTocopy <- readVarNetCDF(ncdNameOld,var)[,,timeIndexINIT:timeIndexEND]
    }
    
    nstep <- (timeIndexEND - timeIndexINIT) + 1
    
    putVarNetCDF(ncdName,var, varTocopy, 1:nstep, as.POSIXct(strTimeToCopy,tz = "UTC"))
    
  }
  
  nc_close(nc)
  
}

putVarNetCDF <- function(ncdName,var.name, var.matrix, nTime, RunTime) {
  
  ncin <- nc_open(ncdName, write = TRUE)
  
  #time_posixct <- ISOdatetime(1900,1,1,0,0,0,tz="UTC")
  time_posixct <- as.POSIXct(RunTime, tz="UTC")
  time_unit <- "hours since 1900-01-01 00:00:00 +00:00"
  time_coord <- utinvcal.nc(time_unit, time_posixct)
  str <- strftime(RunTime,"%Y-%m-%d %H:%M:%S", tz="UTC")
  
  ncvar_put(ncin, "time", time_coord, start=min(nTime), count=length(nTime))
  ncvar_put(ncin, "strTime", str, start=c(1,min(nTime)), count=c(-1,length(nTime)))
  # if (!all(is.na(var.matrix))) ncvar_put(ncin, var.name, var.matrix, start=c(1,1,nTime), count=c(-1,-1,1))
  ncvar_put(ncin, var.name, var.matrix, start=c(1,1,min(nTime)), count=c(-1,-1,length(nTime)))
  
  nc_close(ncin)
}


readVarNetCDF <- function(ncdName,varName) {
  
  ncin <- nc_open(ncdName)
  r <- ncvar_get(ncin, varName)
  nc_close(ncin)
  
  return(r)
  
}

##############################################################
#
#   INIZIO CODICE
#
##############################################################


# converto la griglia in punti
griglia_points <- as.data.frame(rasterToPoints(grigliaUTM))

coordinates(griglia_points) <- ~ x + y
proj4string(griglia_points)=CRS(projutm)


tryCatch( 
  {
  
    if (Sys.info()['sysname']=="Linux") {
      
      # setwd("/opt/R_oracle_linux_8/data/SINERGIE-KRIGING")
      
      args <- commandArgs(trailingOnly = TRUE)
      
      outputDir <- as.character(args[1])
      CSVfiles <- as.character(args[2])
      dirBck <- file.path("/opt/pso/modelli/KRIGING/bck", strftime(Sys.time(), format = "%Y%m%d%H%M_OND"))
      
      
    } else if (Sys.info()['sysname']=="Windows") {

      # TEST ENVIRONMENT
      outputDir <- "output"
      CSVfiles <- "input/ondemand.csv"
      dirBck <- file.path("bck", strftime(Sys.time(), format = "%Y%m%d%H%M"))

    } 
    
    # copia backup
    dir.create(dirBck, showWarnings = FALSE)
    #t <- file.copy(CSVfiles, dirBck, recursive = TRUE)
    # for (file in CSVfiles) {
      t <- file.copy(CSVfiles, dirBck, overwrite = TRUE)
    # } 
    
    for (file in CSVfiles) {
      
      print(strrep("#", 50))
      
      fileName <- sub('\\.csv$', '', basename(path = file)) 
      
      print(paste0(Sys.time()," - Elaboro file: ",fileName))
      
      ########################################
      # KRIGING
      
      pioggia <- read.table(file,header = FALSE, sep = ";", dec = ".", skip = 1)
      names(pioggia) <- c("ID_SENSORE","ID_STAZIONE","UTM_EST","UTM_NORD","VALORE","DT_RILEVAMENTO")
      
      for (dt in rev(unique(pioggia$DT_RILEVAMENTO))) {
        
        pioggiaDT <- pioggia[pioggia$DT_RILEVAMENTO == dt,]
        
        valoriPioggia <- pioggiaDT[pioggiaDT$VALORE >= 0 & !is.na(pioggiaDT$VALORE),]
        
        if ((nrow(valoriPioggia) > 0)&(sum(valoriPioggia$VALORE) > 1)) {
          
          coordinates(valoriPioggia) <- ~ UTM_EST + UTM_NORD
          proj4string(valoriPioggia)=CRS(projutm)
          
          if (is.null(variogram(VALORE ~ 1, valoriPioggia))) {
            
            print(paste0(Sys.time()," - Troppi pochi punti per fare il kriging!"))
            r <- grigliaUTM
            values(r) <- NA
            
          } else {
            
            krigingPoints <- autoKrige(VALORE ~ 1, valoriPioggia, new_data=griglia_points)
            r <- rasterFromXYZ(as.data.frame(krigingPoints$krige_output)[, c("x", "y", "var1.pred")])
            r[r<0] <- 0
            
          }
          
        } else {
          
          r <- grigliaUTM
          values(r) <- NA
          
        }
        
        projection(r) <- projutm
        r.proj <- projectRaster(r, grigliaLL)
        
        r.matrix <- t(as.matrix(r.proj))
        
        #########################################
        
        varName <- "CUM"
        dateNTC <- as.POSIXlt(dt, format="%Y-%m-%d %H:%M:%S", tz="UTC")
        monthYear <- strftime(dateNTC, format="%Y%m")
        strTime <- strftime(dateNTC, format="%Y-%m-%d %H:%M:%S")
        
        # type <- str_split(fileName, "_", simplify = TRUE)[1]
        type <- "ONDEMAND"
        
        print(paste0(Sys.time()," - Elaborazione: ",type))
        print(paste0(Sys.time()," - Istante temporale: ",strTime))
        print(paste0(Sys.time()," - Variabile: ",varName))
        
        
        ### ONDEMAND ###
        
        NETCDFfile <- paste0(outputDir,"/",NTCNameOndemand)
        timestep <- timestepOndemand
        #totalsteps <- totalstepsOndemand 
        NTCVariablesList <- NTCVariablesListOndemand
        UnitsList <- UnitsListOndemand
        LongNameList <- LongNameListOndemand
        
          
        # print(paste0(Sys.time()," - ", NETCDFfile, " non esiste! Lo creo..."))
        
        createNetCDF_LatLon(NETCDFfile,raster::xFromCol(grigliaLL),raster::yFromRow(grigliaLL),NTCVariablesList,UnitsList,LongNameList)
        nStep <- 1
        
        putVarNetCDF(NETCDFfile,varName,r.matrix,nStep,dateNTC)
        
        print(paste0(Sys.time()," - Timestep inserito: ",nStep, " - ", dateNTC))
        
      }
      
    }
    
    write.table("OK", paste0(outputDir,"/resultOND.log"), col.names = F, row.names = F,quote = F)
  
  },
  error=function(cond) {
    
    print(paste0(Sys.time()," - ERRORE: "))
    print(paste0(Sys.time(), cond))
    
    write.table("ERROR", paste0(outputDir,"/resultOND.log"), col.names = F, row.names = F,quote = F)
    
  }
)


